@@ -54,13 +54,13 @@ GStaticPrivate job_private = G_STATIC_PRIVATE_INIT;
  * Since this id debug only, just use what the macro should be here.
  * even though it is not portable.
  */
-GStaticMutex debug_mutex = { NULL, { { } } };
+GStaticMutex debug_mutex = G_STATIC_MUTEX_INIT;
 #endif
 
 static int job_count = 0;
 
 static void     gnome_vfs_op_destroy                (GnomeVFSOp           *op);
-static void     gnome_vfs_job_destroy_notify_result (GnomeVFSNotifyResult *notify_result);
+static void     _gnome_vfs_job_destroy_notify_result (GnomeVFSNotifyResult *notify_result);
 static gboolean dispatch_job_callback               (gpointer              data);
 static gboolean dispatch_sync_job_callback          (gpointer              data);
 
@@ -107,9 +107,13 @@ clr_fl (int fd, int flags)
 	}
 }
 
-/* Find out whether or not a given job has more ops to go */
+/*
+ *   Find out whether or not a given job should be left in
+ * the job map, preserving it's open VFS handle, since we
+ * can do more operations on it later.
+ */
 gboolean
-gnome_vfs_job_complete (GnomeVFSJob *job)
+_gnome_vfs_job_complete (GnomeVFSJob *job)
 {
 	g_assert (job->op != NULL);
 	
@@ -127,6 +131,7 @@ gnome_vfs_job_complete (GnomeVFSJob *job)
 		g_assert_not_reached();
 		return FALSE;
 	case GNOME_VFS_OP_READ_WRITE_DONE:
+	case GNOME_VFS_OP_FILE_CONTROL:
 		return FALSE;
 	
 	default:
@@ -140,11 +145,16 @@ gnome_vfs_job_complete (GnomeVFSJob *job)
 static void
 job_oneway_notify (GnomeVFSJob *job, GnomeVFSNotifyResult *notify_result)
 {
-	if (gnome_vfs_async_job_add_callback (job, notify_result)) {
+	if (_gnome_vfs_async_job_add_callback (job, notify_result)) {
 		JOB_DEBUG (("job %u, callback %u", GPOINTER_TO_UINT (notify_result->job_handle),
 			notify_result->callback_id));
 	
 		g_idle_add (dispatch_job_callback, notify_result);
+	} else {
+		JOB_DEBUG (("Barfing on oneway cancel %u (%d)",
+			    GPOINTER_TO_UINT (notify_result->job_handle),
+			    job->op->type));
+		_gnome_vfs_job_destroy_notify_result (notify_result);
 	}
 }
 
@@ -153,40 +163,24 @@ job_oneway_notify (GnomeVFSJob *job, GnomeVFSNotifyResult *notify_result)
 static void
 job_notify (GnomeVFSJob *job, GnomeVFSNotifyResult *notify_result)
 {
-	if (!gnome_vfs_async_job_add_callback (job, notify_result)) {
-		JOB_DEBUG (("job cancelled, bailing %u",
-			GPOINTER_TO_UINT (notify_result->job_handle)));
+	if (!_gnome_vfs_async_job_add_callback (job, notify_result)) {
+		JOB_DEBUG (("Barfing on sync cancel %u (%d)",
+			    GPOINTER_TO_UINT (notify_result->job_handle),
+			    job->op->type));
+		_gnome_vfs_job_destroy_notify_result (notify_result);
 		return;
 	}
 
-	JOB_DEBUG (("Locking notification lock %u", GPOINTER_TO_UINT (notify_result->job_handle)));
-	/* Lock notification, so that the master cannot send the signal until
-           we are ready to receive it.  */
-	g_mutex_lock (job->notify_ack_lock);
-
 	/* Send the notification.  This will wake up the master thread, which
          * will in turn signal the notify condition.
          */
 	g_idle_add (dispatch_sync_job_callback, notify_result);
 
-	/* FIXME:
-	 * unlock here to prevent deadlock with async cancel. We should not use the
-	 * access lock at all in the case of synch operations like xfer.
-	 * Unlocking here is perfectly OK, even though it's a hack.
-	 */
-	sem_post (&job->access_lock);
-
 	JOB_DEBUG (("Wait notify condition %u", GPOINTER_TO_UINT (notify_result->job_handle)));
 	/* Wait for the notify condition.  */
-	g_cond_wait (job->notify_ack_condition, job->notify_ack_lock);
+	g_cond_wait (job->notify_ack_condition, job->job_lock);
 
-	sem_wait (&job->access_lock);
-
-	JOB_DEBUG (("Unlock notify ack lock %u", GPOINTER_TO_UINT (notify_result->job_handle)));
-	/* Acknowledgment got: unlock the mutex.  */
-	g_mutex_unlock (job->notify_ack_lock);
-
-	JOB_DEBUG (("Done %u", GPOINTER_TO_UINT (notify_result->job_handle)));
+	JOB_DEBUG (("Got notify ack condition %u", GPOINTER_TO_UINT (notify_result->job_handle)));
 }
 
 static void
@@ -320,6 +314,15 @@ dispatch_module_callback (GnomeVFSNotifyResult *notify_result)
 						    notify_result->specifics.callback.response_data);
 }
 
+static void
+dispatch_file_control_callback (GnomeVFSNotifyResult *notify_result)
+{
+	notify_result->specifics.file_control.callback (notify_result->job_handle,
+							notify_result->specifics.file_control.result,
+							notify_result->specifics.file_control.operation_data,
+							notify_result->specifics.file_control.callback_data);
+}
+
 static void
 empty_close_callback (GnomeVFSAsyncHandle *handle,
 		      GnomeVFSResult result,
@@ -331,12 +334,11 @@ static void
 handle_cancelled_open (GnomeVFSJob *job)
 {
 	/* schedule a silent close to make sure the handle does not leak */
-	gnome_vfs_job_set (job, GNOME_VFS_OP_CLOSE,
+	_gnome_vfs_job_set (job, GNOME_VFS_OP_CLOSE, 
 			   (GFunc) empty_close_callback, NULL);
-	gnome_vfs_job_go (job);
+	_gnome_vfs_job_go (job);
 }
 
-
 static void
 free_get_file_info_notify_result (GnomeVFSGetFileInfoOpResult *notify_result)
 {
@@ -371,7 +373,7 @@ free_find_directory_notify_result (GnomeVFSFindDirectoryOpResult *notify_result)
 }
 
 static void
-gnome_vfs_job_destroy_notify_result (GnomeVFSNotifyResult *notify_result)
+_gnome_vfs_job_destroy_notify_result (GnomeVFSNotifyResult *notify_result)
 {
 	JOB_DEBUG (("%u", notify_result->callback_id));
 
@@ -384,6 +386,15 @@ gnome_vfs_job_destroy_notify_result (GnomeVFSNotifyResult *notify_result)
 	case GNOME_VFS_OP_OPEN:
 	case GNOME_VFS_OP_OPEN_AS_CHANNEL:
 	case GNOME_VFS_OP_READ:
+	case GNOME_VFS_OP_XFER:
+		g_free (notify_result);
+		break;
+		
+	case GNOME_VFS_OP_FILE_CONTROL:
+		if (notify_result->specifics.file_control.operation_data_destroy_func) {
+			notify_result->specifics.file_control.operation_data_destroy_func (
+					    notify_result->specifics.file_control.operation_data);
+		}
 		g_free (notify_result);
 		break;
 		
@@ -407,6 +418,15 @@ gnome_vfs_job_destroy_notify_result (GnomeVFSNotifyResult *notify_result)
 		g_free (notify_result);
 		break;
 
+	case GNOME_VFS_OP_MODULE_CALLBACK:
+		/* TODO: This seems to happen when you get multiple auth dialogs in at least the
+		 * smb backend. I don't know why, if its right, or what to do here if it is right.
+		 * But if we warn like this at least we won't die in the assert_not_reached, and it
+		 * seems to work. This needs more looking into.
+		 */
+		g_warning ("_gnome_vfs_job_destroy_notify_result(): GNOME_VFS_OP_MODULE_CALLBACK notify_result destroyed. Is this supposed to happen?\n"); 
+		break;
+	
 	default:
 		g_assert_not_reached ();
 		break;
@@ -424,17 +444,23 @@ dispatch_sync_job_callback (gpointer data)
 
 	notify_result = (GnomeVFSNotifyResult *) data;
 
-	gnome_vfs_async_job_callback_valid (notify_result->callback_id, &valid, &cancelled);
+	_gnome_vfs_async_job_callback_valid (notify_result->callback_id, &valid, &cancelled);
 
 	/* Even though the notify result is owned by the async thread and persists
 	 * all through the notification, we still keep it in the job map to
 	 * make cancellation easier.
 	 */
-	gnome_vfs_async_job_remove_callback (notify_result->callback_id);
+	_gnome_vfs_async_job_remove_callback (notify_result->callback_id);
 
 	g_assert (valid);
 
 	switch (notify_result->type) {
+	case GNOME_VFS_OP_CREATE_AS_CHANNEL:
+		dispatch_create_as_channel_callback (notify_result);
+		break;
+	case GNOME_VFS_OP_OPEN_AS_CHANNEL:
+		dispatch_open_as_channel_callback (notify_result);
+		break;		
 	case GNOME_VFS_OP_XFER:
 		dispatch_xfer_callback (notify_result, cancelled);
 		break;
@@ -446,22 +472,18 @@ dispatch_sync_job_callback (gpointer data)
 		break;
 	}
 	
-	gnome_vfs_async_job_map_lock ();
-	job = gnome_vfs_async_job_map_get_job (notify_result->job_handle);
-	gnome_vfs_async_job_map_unlock ();
+	_gnome_vfs_async_job_map_lock ();
+	job = _gnome_vfs_async_job_map_get_job (notify_result->job_handle);
+	g_mutex_lock (job->job_lock);
+	_gnome_vfs_async_job_map_unlock ();
 	
 	g_assert (job != NULL);
 	
 	JOB_DEBUG (("signalling %u", GPOINTER_TO_UINT (notify_result->job_handle)));
-
-	/* OK to access a job with the map unlocked, job will not be deleted
-	 * from under us.
-	 */
-	g_mutex_lock (job->notify_ack_lock);
 	
 	/* Signal the async thread that we are done with the notification. */
 	g_cond_signal (job->notify_ack_condition);
-	g_mutex_unlock (job->notify_ack_lock);
+	g_mutex_unlock (job->job_lock);
 
 	return FALSE;
 }
@@ -480,55 +502,58 @@ dispatch_job_callback (gpointer data)
 
 	JOB_DEBUG (("%u", GPOINTER_TO_UINT (notify_result->job_handle)));	
 	
-	gnome_vfs_async_job_callback_valid (notify_result->callback_id, &valid, &cancelled);
-	gnome_vfs_async_job_remove_callback (notify_result->callback_id);
+	_gnome_vfs_async_job_callback_valid (notify_result->callback_id, &valid, &cancelled);
+	_gnome_vfs_async_job_remove_callback (notify_result->callback_id);
 
 	if (!valid) {
 		/* this can happen when gnome vfs is shutting down */
-		JOB_DEBUG (("callback %u no longer valid", notify_result->callback_id));
-		gnome_vfs_job_destroy_notify_result (notify_result);
+		JOB_DEBUG (("shutting down: callback %u no longer valid",
+			    notify_result->callback_id));
+		_gnome_vfs_job_destroy_notify_result (notify_result);
 		return FALSE;
 	}
 	
 	if (cancelled) {
 		/* cancel the job in progress */
-		
-		JOB_DEBUG (("cancelling job %u %u", GPOINTER_TO_UINT (notify_result->job_handle),
-			notify_result->callback_id));
+		JOB_DEBUG (("cancelling job %u %u",
+			    GPOINTER_TO_UINT (notify_result->job_handle),
+			    notify_result->callback_id));
 
-		gnome_vfs_async_job_map_lock ();
+		_gnome_vfs_async_job_map_lock ();
 
-		job = gnome_vfs_async_job_map_get_job (notify_result->job_handle);
+		job = _gnome_vfs_async_job_map_get_job (notify_result->job_handle);
 		
 		if (job != NULL) {
-			/* If needed, schedule a close to make sure we do not leak open handles. */
+			g_mutex_lock (job->job_lock);
+
 			switch (job->op->type) {
 			case GNOME_VFS_OP_OPEN:
-			case GNOME_VFS_OP_CREATE:
 			case GNOME_VFS_OP_OPEN_AS_CHANNEL:
+			case GNOME_VFS_OP_CREATE:
 			case GNOME_VFS_OP_CREATE_AS_CHANNEL:
-				JOB_DEBUG (("cancelling open or create %u", GPOINTER_TO_UINT (job->job_handle)));
-				handle_cancelled_open (job);
-				
-				/* Keep the job in the job map -- it will get removed once close completes. */
-				break;
-		
+			case GNOME_VFS_OP_CREATE_SYMBOLIC_LINK:
+				if (job->handle) {
+					g_mutex_unlock (job->job_lock);
+					handle_cancelled_open (job);
+					JOB_DEBUG (("handle cancel open job %u",
+						    GPOINTER_TO_UINT (notify_result->job_handle)));
+					break;
+				} /* else drop through */
 			default:
 				/* Remove job from the job map. */
-				gnome_vfs_async_job_map_remove_job (job);
+				_gnome_vfs_async_job_map_remove_job (job);
+				g_mutex_unlock (job->job_lock);
 				break;
 			}
-			
 		}
 	
-		gnome_vfs_async_job_map_unlock ();
-		gnome_vfs_job_destroy_notify_result (notify_result);
+		_gnome_vfs_async_job_map_unlock ();
+		_gnome_vfs_job_destroy_notify_result (notify_result);
 		return FALSE;
 	}
 	
 		
 	JOB_DEBUG (("executing callback %u", GPOINTER_TO_UINT (notify_result->job_handle)));	
-	
 
 	switch (notify_result->type) {
 	case GNOME_VFS_OP_CLOSE:
@@ -567,63 +592,67 @@ dispatch_job_callback (gpointer data)
 	case GNOME_VFS_OP_WRITE:
 		dispatch_write_callback (notify_result);
 		break;
+	case GNOME_VFS_OP_FILE_CONTROL:
+		dispatch_file_control_callback (notify_result);
+		break;
 	default:
 		g_assert_not_reached ();
 		break;
 	}
 
 	JOB_DEBUG (("dispatch callback - done %u", GPOINTER_TO_UINT (notify_result->job_handle)));
-	gnome_vfs_job_destroy_notify_result (notify_result);
+	_gnome_vfs_job_destroy_notify_result (notify_result);
 
 	return FALSE;
 }
 
 void
-gnome_vfs_job_set (GnomeVFSJob *job,
+_gnome_vfs_job_set (GnomeVFSJob *job,
 		   GnomeVFSOpType type,
 		   GFunc callback,
 		   gpointer callback_data)
 {
 	GnomeVFSOp *op;
 
-	JOB_DEBUG (("locking access lock %u, op %d", GPOINTER_TO_UINT (job->job_handle), type));
-	sem_wait (&job->access_lock);
-
 	op = g_new (GnomeVFSOp, 1);
 	op->type = type;
 	op->callback = callback;
 	op->callback_data = callback_data;
 	op->context = gnome_vfs_context_new ();
-	op->stack_info = gnome_vfs_module_callback_get_stack_info ();
+	op->stack_info = _gnome_vfs_module_callback_get_stack_info ();
 
 	g_assert (gnome_vfs_context_get_cancellation (op->context) != NULL);
 
+	JOB_DEBUG (("locking access lock %u, op %d", GPOINTER_TO_UINT (job->job_handle), type));
+
+	g_mutex_lock (job->job_lock);
+
 	gnome_vfs_op_destroy (job->op);
 	job->op = op;
-
 	job->cancelled = FALSE;
 
+	g_mutex_unlock (job->job_lock);
+
 	JOB_DEBUG (("%u op type %d, op %p", GPOINTER_TO_UINT (job->job_handle),
 		job->op->type, job->op));
 }
 
 GnomeVFSJob *
-gnome_vfs_job_new (GnomeVFSOpType type, int priority, GFunc callback, gpointer callback_data)
+_gnome_vfs_job_new (GnomeVFSOpType type, int priority, GFunc callback, gpointer callback_data)
 {
 	GnomeVFSJob *new_job;
 	
 	new_job = g_new0 (GnomeVFSJob, 1);
-	
-	sem_init (&new_job->access_lock, 0, 1);
+
+	new_job->job_lock = g_mutex_new ();
 	new_job->notify_ack_condition = g_cond_new ();
-	new_job->notify_ack_lock = g_mutex_new ();
 	new_job->priority = priority;
 
 	/* Add the new job into the job hash table. This also assigns
 	 * the job a unique id
 	 */
-	gnome_vfs_async_job_map_add_job (new_job);
-	gnome_vfs_job_set (new_job, type, callback, callback_data);
+	_gnome_vfs_async_job_map_add_job (new_job);
+	_gnome_vfs_job_set (new_job, type, callback, callback_data);
 
 	job_count++;
 
@@ -631,16 +660,16 @@ gnome_vfs_job_new (GnomeVFSOpType type, int priority, GFunc callback, gpointer c
 }
 
 void
-gnome_vfs_job_destroy (GnomeVFSJob *job)
+_gnome_vfs_job_destroy (GnomeVFSJob *job)
 {
 	JOB_DEBUG (("destroying job %u", GPOINTER_TO_UINT (job->job_handle)));
 
 	gnome_vfs_op_destroy (job->op);
 
-	sem_destroy (&job->access_lock);
-
+	g_mutex_free (job->job_lock);
 	g_cond_free (job->notify_ack_condition);
-	g_mutex_free (job->notify_ack_lock);
+
+	memset (job, 0xaa, sizeof (GnomeVFSJob));
 
 	g_free (job);
 	job_count--;
@@ -710,32 +739,33 @@ gnome_vfs_op_destroy (GnomeVFSOp *op)
 	case GNOME_VFS_OP_CLOSE:
 	case GNOME_VFS_OP_READ_WRITE_DONE:
 		break;
+	case GNOME_VFS_OP_FILE_CONTROL:
+		g_free (op->specifics.file_control.operation);
+		break;
 	default:
 		g_warning (_("Unknown op type %u"), op->type);
 	}
 	
 	g_assert (gnome_vfs_context_get_cancellation (op->context) != NULL);
 	
-	gnome_vfs_context_unref (op->context);
-	gnome_vfs_module_callback_free_stack_info (op->stack_info);
+	gnome_vfs_context_free (op->context);
+	_gnome_vfs_module_callback_free_stack_info (op->stack_info);
 	
 	g_free (op);
 }
 
 void
-gnome_vfs_job_go (GnomeVFSJob *job)
+_gnome_vfs_job_go (GnomeVFSJob *job)
 {
-	/* Fire up the async job thread. */
-	if (!gnome_vfs_job_schedule (job)) {
-		g_warning ("Cannot schedule this job.");
-		gnome_vfs_job_destroy (job);
-		return;
-	}
-	
-	JOB_DEBUG (("new job %u, op %d, unlocking access lock",
+	JOB_DEBUG (("new job %u, op %d, unlocking job lock",
 		GPOINTER_TO_UINT (job->job_handle), job->op->type));
 
-	sem_post (&job->access_lock);
+	/* Fire up the async job thread. */
+	if (!_gnome_vfs_job_schedule (job)) {
+		g_warning ("Cannot schedule this job.");
+		_gnome_vfs_job_destroy (job);
+		return;
+	}
 }
 
 #define DEFAULT_BUFFER_SIZE 16384
@@ -763,7 +793,7 @@ serve_channel_read (GnomeVFSHandle *handle,
 
 	while (1) {
 		GnomeVFSResult result;
-		GIOError io_result;
+		GIOStatus io_result;
 		GnomeVFSFileSize bytes_read;
 		
 	restart_toplevel_loop:
@@ -779,7 +809,7 @@ serve_channel_read (GnomeVFSHandle *handle,
 
 		if (result == GNOME_VFS_ERROR_INTERRUPTED) {
 			continue;
-		} else if (result != GNOME_VFS_OK) {
+		} else if (result != GNOME_VFS_OK && result != GNOME_VFS_ERROR_EOF) {
 			goto end;
 		}
 	
@@ -799,7 +829,7 @@ serve_channel_read (GnomeVFSHandle *handle,
 			gsize bytes_written;
 			
 			/* channel_out is nonblocking; if we get
-			   EAGAIN (G_IO_ERROR_AGAIN) then we tried to
+			   EAGAIN (G_IO_STATUS_AGAIN) then we tried to
 			   write but the pipe was full. In this case, we
 			   want to enlarge our buffer and go back to
 			   reading for one iteration, so we can keep
@@ -812,11 +842,13 @@ serve_channel_read (GnomeVFSHandle *handle,
 				 filled_bytes_in_buffer - written_bytes_in_buffer,
 				 &bytes_written, NULL);
 			
+			written_bytes_in_buffer += bytes_written;
+
 			if (gnome_vfs_context_check_cancellation(context)) {
 				goto end;
 			}
 			
-			if (io_result == G_IO_ERROR_AGAIN) {
+			if (io_result == G_IO_STATUS_AGAIN) {
 				/* if bytes_read == 0 then we reached
 				   EOF so there's no point reading
 				   again. So turn off nonblocking and
@@ -852,11 +884,9 @@ serve_channel_read (GnomeVFSHandle *handle,
 
 				} /* end of else (bytes_read != 0) */
 				
-			} else if (io_result != G_IO_ERROR_NONE || bytes_written == 0) {
+			} else if (io_result != G_IO_STATUS_NORMAL || bytes_written == 0) {
 				goto end;
 			}
-
-			written_bytes_in_buffer += bytes_written;
 		}
 
 		g_assert(written_bytes_in_buffer == filled_bytes_in_buffer);
@@ -886,7 +916,7 @@ serve_channel_write (GnomeVFSHandle *handle,
 
 	while (1) {
 		GnomeVFSResult result;
-		GIOError io_result;
+		GIOStatus io_result;
 		gsize bytes_read;
 		gsize bytes_to_write;
 		GnomeVFSFileSize bytes_written;
@@ -894,10 +924,10 @@ serve_channel_write (GnomeVFSHandle *handle,
 
 		io_result = g_io_channel_read_chars (channel_in, buffer, buffer_size,
 						     &bytes_read, NULL);
-		if (io_result == G_IO_ERROR_AGAIN || io_result == G_IO_ERROR_UNKNOWN)
-			/* we will get G_IO_ERROR_UNKNOWN if a signal occurrs */
+
+		if (io_result == G_IO_STATUS_AGAIN)
 			continue;
-		if (io_result != G_IO_ERROR_NONE || bytes_read == 0)
+		if (io_result != G_IO_STATUS_NORMAL || bytes_read == 0)
 			goto end;
 
 		p = buffer;
@@ -1030,7 +1060,7 @@ execute_open_as_channel (GnomeVFSJob *job)
 
 	notify_result->specifics.open_as_channel.result = GNOME_VFS_OK;
 
-	job_oneway_notify (job, notify_result);
+	job_notify (job, notify_result);
 
 	if (open_mode & GNOME_VFS_OPEN_READ) {
 		serve_channel_read (handle, channel_in, channel_out,
@@ -1161,7 +1191,7 @@ execute_create_as_channel (GnomeVFSJob *job)
 
 	notify_result->specifics.create_as_channel.channel = channel_out;
 
-	job_oneway_notify (job, notify_result);
+	job_notify (job, notify_result);
 
 	serve_channel_write (handle, channel_in, channel_out, job->op->context);
 }
@@ -1207,6 +1237,8 @@ execute_read (GnomeVFSJob *job)
 									   &notify_result->specifics.read.bytes_read,
 									   job->op->context);
 
+	job->op->type = GNOME_VFS_OP_READ_WRITE_DONE;
+
 	job_oneway_notify (job, notify_result);
 }
 
@@ -1232,6 +1264,7 @@ execute_write (GnomeVFSJob *job)
 									     &notify_result->specifics.write.bytes_written,
 									     job->op->context);
 
+	job->op->type = GNOME_VFS_OP_READ_WRITE_DONE;
 
 	job_oneway_notify (job, notify_result);
 }
@@ -1401,7 +1434,6 @@ load_directory_details (GnomeVFSJob *job)
 			(GnomeVFSAsyncDirectoryLoadCallback) job->op->callback;
 		notify_result->specifics.load_directory.callback_data = job->op->callback_data;
 		job_oneway_notify (job, notify_result);
-		
 		return;
 	}
 
@@ -1410,16 +1442,18 @@ load_directory_details (GnomeVFSJob *job)
 	count = 0;
 	while (1) {
 		if (gnome_vfs_context_check_cancellation (job->op->context)) {
-			JOB_DEBUG (("cancelled, bailing %u", GPOINTER_TO_UINT (job->job_handle)));
+			JOB_DEBUG (("cancelled, bailing %u",
+				    GPOINTER_TO_UINT (job->job_handle)));
 			gnome_vfs_file_info_list_free (directory_list);
 			directory_list = NULL;
 			result = GNOME_VFS_ERROR_CANCELLED;
 			break;
 		}
-		
+
 		info = gnome_vfs_file_info_new ();
 
-		result = gnome_vfs_directory_read_next_cancellable (handle, info, job->op->context);
+		result = gnome_vfs_directory_read_next_cancellable
+			(handle, info, job->op->context);
 
 		if (result == GNOME_VFS_OK) {
 			directory_list = g_list_prepend (directory_list, info);
@@ -1502,7 +1536,7 @@ execute_xfer (GnomeVFSJob *job)
 
 	xfer_op = &job->op->specifics.xfer;
 
-	result = gnome_vfs_xfer_private (xfer_op->source_uri_list,
+	result = _gnome_vfs_xfer_private (xfer_op->source_uri_list,
 					 xfer_op->target_uri_list,
 					 xfer_op->xfer_options,
 					 xfer_op->error_mode,
@@ -1541,17 +1575,54 @@ execute_xfer (GnomeVFSJob *job)
 	}
 }
 
-/* This function is called by the slave thread to execute a
-   GnomeVFSJob.  */
-void
-gnome_vfs_job_execute (GnomeVFSJob *job)
+static void
+execute_file_control (GnomeVFSJob *job)
 {
-	JOB_DEBUG (("%u", GPOINTER_TO_UINT (job->job_handle)));
+	GnomeVFSFileControlOp *file_control_op;
+	GnomeVFSNotifyResult *notify_result;
+	
+	file_control_op = &job->op->specifics.file_control;
+
+	notify_result = g_new0 (GnomeVFSNotifyResult, 1);
+	notify_result->job_handle = job->job_handle;
+	notify_result->type = job->op->type;
+	notify_result->specifics.file_control.callback = (GnomeVFSAsyncFileControlCallback) job->op->callback;
+	notify_result->specifics.file_control.callback_data = job->op->callback_data;
+	notify_result->specifics.file_control.operation_data = file_control_op->operation_data;
+	notify_result->specifics.file_control.operation_data_destroy_func = file_control_op->operation_data_destroy_func;
+	
+	notify_result->specifics.file_control.result = gnome_vfs_file_control_cancellable (job->handle,
+											   file_control_op->operation,
+											   file_control_op->operation_data,
+											   job->op->context);
+
+	job->op->type = GNOME_VFS_OP_FILE_CONTROL;
+
+	job_oneway_notify (job, notify_result);
+}
+
+
+/*
+ * _gnome_vfs_job_execute:
+ * @job: the job to execute
+ * 
+ *   This function is called by the slave thread to execute
+ * the job - all work performed by a thread starts here.
+ */
+void
+_gnome_vfs_job_execute (GnomeVFSJob *job)
+{
+	guint id;
+
+	id = GPOINTER_TO_UINT (job->job_handle);
+
+	JOB_DEBUG (("exec job %u", id));
 
 	if (!job->cancelled) {
 		set_current_job (job);
 
-		JOB_DEBUG (("executing %u %d", GPOINTER_TO_UINT (job->job_handle), job->op->type));
+		JOB_DEBUG (("executing %u %d", id, job->op->type));
+
 		switch (job->op->type) {
 		case GNOME_VFS_OP_OPEN:
 			execute_open (job);
@@ -1592,31 +1663,31 @@ gnome_vfs_job_execute (GnomeVFSJob *job)
 		case GNOME_VFS_OP_SET_FILE_INFO:
 			execute_set_file_info (job);
 			break;
+		case GNOME_VFS_OP_FILE_CONTROL:
+			execute_file_control (job);
+			break;
 		default:
 			g_warning (_("Unknown job kind %u"), job->op->type);
 			break;
 		}
-
-
+		/* NB. 'job' is quite probably invalid now */
 		clear_current_job ();
-
-	}
-	
-	switch (job->op->type) {
+	} else {
+		switch (job->op->type) {
 		case GNOME_VFS_OP_READ:
 		case GNOME_VFS_OP_WRITE:
 			job->op->type = GNOME_VFS_OP_READ_WRITE_DONE;
 			break;
-
 		default:
 			break;
+		}
 	}
 	
-	JOB_DEBUG (("done %u", GPOINTER_TO_UINT (job->job_handle)));
+	JOB_DEBUG (("done job %u", id));
 }
 
 void
-gnome_vfs_job_module_cancel (GnomeVFSJob *job)
+_gnome_vfs_job_module_cancel (GnomeVFSJob *job)
 {
 	GnomeVFSCancellation *cancellation;
 
@@ -1646,8 +1717,8 @@ set_current_job (GnomeVFSJob *job)
 
 	g_static_private_set (&job_private, job, NULL);
 
-	gnome_vfs_module_callback_use_stack_info (job->op->stack_info);
-	gnome_vfs_module_callback_set_in_async_thread (TRUE);
+	_gnome_vfs_module_callback_use_stack_info (job->op->stack_info);
+	_gnome_vfs_module_callback_set_in_async_thread (TRUE);
 }
 
 static void
@@ -1655,11 +1726,11 @@ clear_current_job (void)
 {
 	g_static_private_set (&job_private, NULL, NULL);
 
-	gnome_vfs_module_callback_clear_stacks ();
+	_gnome_vfs_module_callback_clear_stacks ();
 }
 
 void
-gnome_vfs_get_current_context (GnomeVFSContext **context)
+_gnome_vfs_get_current_context (GnomeVFSContext **context)
 {
 	GnomeVFSJob *job;
 	
@@ -1675,9 +1746,9 @@ gnome_vfs_get_current_context (GnomeVFSContext **context)
 }
 
 void
-gnome_vfs_dispatch_module_callback (GnomeVFSAsyncModuleCallback callback,
-				    gconstpointer in, size_t in_size,
-				    gpointer out, size_t out_size,
+_gnome_vfs_dispatch_module_callback (GnomeVFSAsyncModuleCallback callback,
+				    gconstpointer in, gsize in_size,
+				    gpointer out, gsize out_size,
 				    gpointer user_data,
 				    GnomeVFSModuleCallbackResponse response,
 				    gpointer response_data)
@@ -1708,4 +1779,3 @@ gnome_vfs_dispatch_module_callback (GnomeVFSAsyncModuleCallback callback,
 
 	job_notify (job, &notify_result);
 }
-
