@@ -62,6 +62,7 @@ G_LOCK_DEFINE_STATIC (fam_connection);
 typedef struct {
 	FAMRequest request;
 	GnomeVFSURI *uri;
+	gboolean     cancelled;
 } FileMonitorHandle;
 
 #endif
@@ -105,7 +106,7 @@ GET_PATH_MAX (void)
 #endif
 
 static gchar *
-get_path_from_uri (GnomeVFSURI *uri)
+get_path_from_uri (GnomeVFSURI const *uri)
 {
 	gchar *path;
 
@@ -125,7 +126,7 @@ get_path_from_uri (GnomeVFSURI *uri)
 }
 
 static gchar *
-get_base_from_uri (GnomeVFSURI *uri)
+get_base_from_uri (GnomeVFSURI const *uri)
 {
 	gchar *escaped_base, *base;
 
@@ -189,8 +190,8 @@ do_open (GnomeVFSMethod *method,
 	}
 
 	if (! (mode & GNOME_VFS_OPEN_RANDOM) && (mode & GNOME_VFS_OPEN_WRITE))
-		mode |= O_TRUNC;
-
+		unix_mode |= O_TRUNC;
+	
 	file_name = get_path_from_uri (uri);
 	if (file_name == NULL)
 		return GNOME_VFS_ERROR_INVALID_URI;
@@ -581,6 +582,7 @@ read_link (const gchar *full_name)
 
                 read_size = readlink (full_name, buffer, size);
 		if (read_size < 0) {
+			g_free (buffer);
 			return NULL;
 		}
                 if (read_size < size) {
@@ -592,6 +594,28 @@ read_link (const gchar *full_name)
 	}
 }
 
+static void
+get_access_info (GnomeVFSFileInfo *file_info,
+              const gchar *full_name)
+{
+     /* FIXME: should check errno after calling access because we don't
+      * want to set valid_fields if something bad happened during one
+      * of the access calls
+      */
+     if (access (full_name, R_OK) == 0) {
+             file_info->permissions |= GNOME_VFS_PERM_ACCESS_READABLE;
+     }
+
+     if (access (full_name, W_OK) == 0) {
+             file_info->permissions |= GNOME_VFS_PERM_ACCESS_WRITABLE;
+     }
+
+     if (access (full_name, X_OK) == 0) {
+             file_info->permissions |= GNOME_VFS_PERM_ACCESS_EXECUTABLE;
+     }
+     file_info->valid_fields |= GNOME_VFS_FILE_INFO_FIELDS_ACCESS;
+}
+
 static GnomeVFSResult
 get_stat_info (GnomeVFSFileInfo *file_info,
 	       const gchar *full_name,
@@ -666,7 +690,7 @@ get_stat_info (GnomeVFSFileInfo *file_info,
 			                /* if we had an earlier ELOOP, don't get in an infinite loop here */
 			        || recursive
 					/* we don't care to follow links */
-				|| lstat (file_info->symlink_name, statptr) != 0
+				|| lstat (symlink_name, statptr) != 0
 					/* we can't make out where this points to */
 				|| !S_ISLNK (statptr->st_mode)) {
 					/* the next level is not a link */
@@ -748,6 +772,10 @@ do_close_directory (GnomeVFSMethod *method,
 	return GNOME_VFS_OK;
 }
 
+#ifndef HAVE_READDIR_R
+G_LOCK_DEFINE_STATIC (readdir);
+#endif
+
 static GnomeVFSResult
 do_read_directory (GnomeVFSMethod *method,
 		   GnomeVFSMethodHandle *method_handle,
@@ -762,6 +790,7 @@ do_read_directory (GnomeVFSMethod *method,
 	handle = (DirectoryHandle *) method_handle;
 	
 	errno = 0;
+#ifdef HAVE_READDIR_R	
 	if (readdir_r (handle->dir, handle->current_entry, &result) != 0) {
 		/* Work around a Solaris bug.
 		 * readdir64_r returns -1 instead of 0 at EOF.
@@ -771,6 +800,19 @@ do_read_directory (GnomeVFSMethod *method,
 		}
 		return gnome_vfs_result_from_errno ();
 	}
+#else
+	G_LOCK (readdir);
+	errno = 0;
+	result = readdir (handle->dir);
+
+	if (result == NULL && errno != 0) {
+		return gnome_vfs_result_from_errno ();
+	}
+	if (result != NULL) {
+		memcpy (handle->current_entry, result, sizeof (struct dirent));
+	}
+	G_UNLOCK (readdir);
+#endif
 	
 	if (result == NULL) {
 		return GNOME_VFS_ERROR_EOF;
@@ -822,6 +864,10 @@ do_get_file_info (GnomeVFSMethod *method,
 		return result;
 	}
 
+	if (options & GNOME_VFS_FILE_INFO_GET_ACCESS_RIGHTS) {
+		get_access_info (file_info, full_name);
+	}
+
 	if (options & GNOME_VFS_FILE_INFO_GET_MIME_TYPE) {
 		get_mime_type (file_info, full_name, options, &statbuf);
 	}
@@ -871,15 +917,43 @@ do_get_file_info_from_handle (GnomeVFSMethod *method,
 	return GNOME_VFS_OK;
 }
 
+GHashTable *fstype_hash = NULL;
+G_LOCK_DEFINE_STATIC (fstype_hash);
+extern char *filesystem_type (char *path, char *relpath, struct stat *statp);
 
 static gboolean
 do_is_local (GnomeVFSMethod *method,
 	     const GnomeVFSURI *uri)
 {
+	gchar *path;
+	gpointer local = NULL;
+
 	g_return_val_if_fail (uri != NULL, FALSE);
 
-	/* We are always a native filesystem.  */
-	return TRUE;
+	path = get_path_from_uri (uri);
+	if (path == NULL)
+		return TRUE; /* GNOME_VFS_ERROR_INVALID_URI */
+
+	G_LOCK (fstype_hash);
+	if (fstype_hash == NULL)
+		fstype_hash = g_hash_table_new_full (
+			g_str_hash, g_str_equal, g_free, NULL);
+	else
+		local = g_hash_table_lookup (fstype_hash, path);
+
+	if (local == NULL) {
+		struct stat statbuf;
+		if (stat (path, &statbuf) == 0) {
+			char *type = filesystem_type (path, path, &statbuf);
+			gboolean is_local = strcmp (type, "nfs") && strcmp (type, "afs");
+			local = GINT_TO_POINTER (is_local ? 1 : -1);
+			g_hash_table_insert (fstype_hash, path, local);
+		}
+	} else
+		g_free (path);
+
+	G_UNLOCK (fstype_hash);
+	return GPOINTER_TO_INT (local) > 0;
 }
 
 
@@ -1102,6 +1176,7 @@ find_trash_in_hierarchy (const char *start_dir, dev_t near_device_id, GnomeVFSCo
 }
 
 static GList *cached_trash_directories;
+G_LOCK_DEFINE_STATIC (cached_trash_directories);
 
 /* Element used to store chached Trash entries in the local, in-memory Trash item cache. */
 typedef struct {
@@ -1148,8 +1223,9 @@ try_creating_trash_in (const char *path, guint permissions)
 }
 
 static char *
-find_disk_top_directory (const char *item_on_disk, dev_t near_device_id,
-	GnomeVFSContext *context)
+find_disk_top_directory (const char *item_on_disk,
+			 dev_t near_device_id,
+			 GnomeVFSContext *context)
 {
 	char *disk_top_directory;
 	struct stat stat_buffer;
@@ -1280,7 +1356,6 @@ update_one_cached_trash_entry (gpointer element, gpointer cast_to_context)
 	}
 }
 
-
 static void
 add_local_cached_trash_entry (dev_t near_device_id, const char *trash_path, const char *mount_point)
 {
@@ -1334,7 +1409,7 @@ read_saved_cached_trash_entries (void)
 {
 	char *cache_file_path;
 	FILE *cache_file;
-	char buffer[2 * PATH_MAX + 1];
+	char buffer[2048];
 	char escaped_mount_point[PATH_MAX], escaped_trash_path[PATH_MAX];
 	char *mount_point, *trash_path;
 	struct stat stat_buffer;
@@ -1352,7 +1427,7 @@ read_saved_cached_trash_entries (void)
 
 	if (cache_file != NULL) {
 		for (;;) {
-			if (fgets(buffer, 2048, cache_file) == NULL) {
+			if (fgets (buffer, sizeof (buffer), cache_file) == NULL) {
 				break;
 			}
 
@@ -1523,14 +1598,19 @@ find_or_create_trash_near (const char *full_name_near, dev_t near_device_id,
 
 /* Find or create a trash directory on the same disk as @full_name_near. Check
  * the local and file cache for matching Trash entries first.
+ *
+ *     This is the only entry point for the trash cache code,
+ * we holds the lock while operating on it only here.
  */
 static char *
 find_trash_directory (const char *full_name_near, dev_t near_device_id, 
-	gboolean create_if_needed, gboolean find_if_needed,
-	guint permissions, GnomeVFSContext *context)
+		      gboolean create_if_needed, gboolean find_if_needed,
+		      guint permissions, GnomeVFSContext *context)
 {
 	char *result;
 
+	G_LOCK (cached_trash_directories);
+
 	/* look in the saved trash locations first */
 	result = find_cached_trash_entry_for_device (near_device_id, find_if_needed);
 
@@ -1568,6 +1648,8 @@ find_trash_directory (const char *full_name_near, dev_t near_device_id,
 		g_free (result);
 		result = NULL;
 	}
+
+	G_UNLOCK (cached_trash_directories);
 	
 	return result;
 }
@@ -1630,6 +1712,11 @@ do_find_directory (GnomeVFSMethod *method,
 
 	switch (kind) {
 	case GNOME_VFS_DIRECTORY_KIND_TRASH:
+		/* Use 0700 (S_IRWXU) for the permissions,
+		 * regardless of the requested permissions, so other
+		 * users can't view the trash files.
+		 */
+		permissions = S_IRWXU;	
 		if (near_item_stat.st_dev != home_volume_stat.st_dev) {
 			/* This volume does not contain our home, we have to find/create the Trash
 			 * elsewhere on the volume. Use a heuristic to find a good place.
@@ -1994,30 +2081,24 @@ do_set_file_info (GnomeVFSMethod *method,
 
 #ifdef HAVE_FAM
 static gboolean
-fam_callback (GIOChannel *source,
-	      GIOCondition condition,
-	      gpointer data)
+fam_do_iter_unlocked (void)
 {
-	FileMonitorHandle *handle;
-	GnomeVFSURI *info_uri = NULL;
-	GnomeVFSMonitorEventType event_type = GNOME_VFS_MONITOR_EVENT_CHANGED;
-
-	G_LOCK (fam_connection);
-
 	while (FAMPending(fam_connection)) {
 		FAMEvent ev;
+		FileMonitorHandle *handle;
+		gboolean cancelled;
+		GnomeVFSMonitorEventType event_type;
+
 		if (FAMNextEvent(fam_connection, &ev) != 1) {
 			FAMClose(fam_connection);
 			g_free(fam_connection);
 			fam_connection = FALSE;
-			G_UNLOCK (fam_connection);
 			return FALSE;
 		}
 
 		handle = (FileMonitorHandle *)ev.userdata;
-
-		info_uri = gnome_vfs_uri_append_file_name (handle->uri, 
-				ev.filename);
+		cancelled = handle->cancelled;
+		event_type = -1;
 
 		switch (ev.code) {
 			case FAMChanged:
@@ -2027,35 +2108,67 @@ fam_callback (GIOChannel *source,
 				event_type = GNOME_VFS_MONITOR_EVENT_DELETED;
 				break;
 			case FAMStartExecuting:
-				event_type = 
-					GNOME_VFS_MONITOR_EVENT_STARTEXECUTING;
+				event_type = GNOME_VFS_MONITOR_EVENT_STARTEXECUTING;
 				break;
 			case FAMStopExecuting:
-				event_type = 
-					GNOME_VFS_MONITOR_EVENT_STOPEXECUTING;
+				event_type = GNOME_VFS_MONITOR_EVENT_STOPEXECUTING;
 				break;
 			case FAMCreated:
 				event_type = GNOME_VFS_MONITOR_EVENT_CREATED;
 				break;
 			case FAMAcknowledge:
+				if (handle->cancelled) {
+					gnome_vfs_uri_unref (handle->uri);
+					g_free (handle);
+				}
+				break;
 			case FAMExists:
 			case FAMEndExist:
 			case FAMMoved:
-				event_type = -1;
+				/* Not supported */
 				break;
 		}
 
+		if (event_type != -1 && !cancelled) {
+			GnomeVFSURI *info_uri;
+			gchar *info_str;
 
-		if (event_type != -1) {
-			gnome_vfs_monitor_callback (
-					(GnomeVFSMethodHandle *)handle,
-					info_uri, event_type);
+			/* 
+			 * FAM can send events with either a absolute or
+			 * relative (from the monitored URI) path, so check if
+			 * the filename starts with '/'.  
+			 */
+			if (ev.filename[0] == '/') {
+				info_str = gnome_vfs_get_uri_from_local_path (ev.filename);
+				info_uri = gnome_vfs_uri_new (info_str);
+				g_free (info_str);
+			} else
+				info_uri = gnome_vfs_uri_append_file_name (handle->uri, ev.filename);
+
+			/* This queues an idle, so there are no reentrancy issues */
+			gnome_vfs_monitor_callback ((GnomeVFSMethodHandle *)handle,
+						    info_uri, 
+						    event_type);
+			gnome_vfs_uri_unref (info_uri);
 		}
 	}
 
+	return TRUE;
+}
+
+static gboolean
+fam_callback (GIOChannel *source,
+	      GIOCondition condition,
+	      gpointer data)
+{
+	gboolean res;
+	G_LOCK (fam_connection);
+
+	res = fam_do_iter_unlocked ();
+
 	G_UNLOCK (fam_connection);
 
-	return TRUE;
+	return res;
 }
 
 
@@ -2071,7 +2184,9 @@ monitor_setup (void)
 	if (fam_connection == NULL) {
 		fam_connection = g_malloc0(sizeof(FAMConnection));
 		if (FAMOpen2(fam_connection, "test-monitor") != 0) {
+#ifdef DEBUG_FAM
 			g_print ("FAMOpen failed, FAMErrno=%d\n", FAMErrno);
+#endif
 			g_free(fam_connection);
 			fam_connection = NULL;
 			G_UNLOCK (fam_connection);
@@ -2106,8 +2221,13 @@ do_monitor_add (GnomeVFSMethod *method,
 
 	handle = g_new0 (FileMonitorHandle, 1);
 	handle->uri = uri;
+	handle->cancelled = FALSE;
 	gnome_vfs_uri_ref (uri);
 	filename = get_path_from_uri (uri);
+
+	/* We need to queue up incoming messages to avoid blocking on write
+	   if there are many monitors being added */
+	fam_do_iter_unlocked ();
 	
 	if (monitor_type == GNOME_VFS_MONITOR_FILE) {
 		G_LOCK (fam_connection);
@@ -2142,20 +2262,39 @@ do_monitor_cancel (GnomeVFSMethod *method,
 		return GNOME_VFS_ERROR_NOT_SUPPORTED;
 	}
 
+	if (handle->cancelled)
+		return GNOME_VFS_OK;
+
+	handle->cancelled = TRUE;
 	G_LOCK (fam_connection);
+
+	/* We need to queue up incoming messages to avoid blocking on write
+	   if there are many monitors being canceled */
+	fam_do_iter_unlocked ();
+	
 	FAMCancelMonitor (fam_connection, &handle->request);
 	G_UNLOCK (fam_connection);
 
-	gnome_vfs_uri_unref (handle->uri);
-
-	g_free (handle);
-
 	return GNOME_VFS_OK;
 #else
 	return GNOME_VFS_ERROR_NOT_SUPPORTED;
 #endif
 }
 
+static GnomeVFSResult
+do_file_control (GnomeVFSMethod *method,
+		 GnomeVFSMethodHandle *method_handle,
+		 const char *operation,
+		 gpointer operation_data,
+		 GnomeVFSContext *context)
+{
+	if (strcmp (operation, "file:test") == 0) {
+		*(char **)operation_data = g_strdup ("test ok");
+		return GNOME_VFS_OK;
+	}
+	return GNOME_VFS_ERROR_NOT_SUPPORTED;
+}
+
 static GnomeVFSMethod method = {
 	sizeof (GnomeVFSMethod),
 	do_open,
@@ -2182,7 +2321,8 @@ static GnomeVFSMethod method = {
 	do_find_directory,
 	do_create_symbolic_link,
 	do_monitor_add,
-	do_monitor_cancel
+	do_monitor_cancel,
+	do_file_control
 };
 
 GnomeVFSMethod *
