@@ -37,8 +37,13 @@
 #include <unistd.h>
 #include <signal.h>
 #include <sys/signal.h>
+#include <fcntl.h>
 
-#define LINE_LENGTH 4096 /* max line length we'll grok */
+#define D(x)
+/* #define D(x) g_print x */
+
+#define LINE_LENGTH 4096	/* max line length we'll grok */
+#define SLEEP_TIME 300000	/* time we'll wait for the process to finish */
 
 typedef struct {
 	GnomeVFSMethodHandle method_handle;
@@ -50,9 +55,15 @@ typedef struct {
 	GnomeVFSOpenMode open_mode;
 	int read_fd;
 	int write_fd;
+	int error_fd;
 	pid_t pid;
+	GnomeVFSFileInfoOptions info_opts;
 } SshHandle;
 
+static GnomeVFSResult ssh_read_error	(int error_fd,
+					 gpointer buffer,
+					 GnomeVFSFileSize num_bytes,
+					 GnomeVFSFileSize *bytes_read);
 static GnomeVFSResult do_open           (GnomeVFSMethod *method,
 				         GnomeVFSMethodHandle **method_handle,
 				         GnomeVFSURI *uri,
@@ -104,9 +115,19 @@ static GnomeVFSResult do_make_directory (GnomeVFSMethod *method,
 static GnomeVFSResult do_remove_directory(GnomeVFSMethod *method,
 					  GnomeVFSURI *uri,
 					  GnomeVFSContext *context);
+static GnomeVFSResult do_move            (GnomeVFSMethod *method,
+					  GnomeVFSURI *old_uri,
+					  GnomeVFSURI *new_uri,
+					  gboolean force_replace,
+					  GnomeVFSContext *context);
 static GnomeVFSResult do_unlink         (GnomeVFSMethod *method,
 					 GnomeVFSURI *uri,
 					 GnomeVFSContext *context);
+static GnomeVFSResult do_check_same_fs	(GnomeVFSMethod  *method,
+					 GnomeVFSURI     *a,
+					 GnomeVFSURI     *b,
+					 gboolean        *same_fs_return,
+					 GnomeVFSContext *context);
 static GnomeVFSResult do_set_file_info  (GnomeVFSMethod *method,
 					 GnomeVFSURI *uri,
 					 const GnomeVFSFileInfo *info,
@@ -139,9 +160,9 @@ static GnomeVFSMethod method = {
         do_is_local,
 	do_make_directory, /* make directory */
         do_remove_directory, /* remove directory */
-	NULL, /* move */
+	do_move,
 	do_unlink, /* unlink */
-	NULL, /* check_same_fs */
+	do_check_same_fs,
 	do_set_file_info, /* set_file_info */
 	NULL, /* truncate */
 	NULL, /* find_directory */
@@ -155,51 +176,89 @@ ssh_connect (SshHandle **handle_return,
 {
 	char ** argv;
 	SshHandle *handle;
-	char *command_line;
-	const gchar *username;
+	char *command_line, *host_port;
+	const gchar *username, *hostname;
 	int argc;
 	GError *gerror = NULL;
+	GnomeVFSFileSize bytes_read;
+	char buffer[LINE_LENGTH];
+	GnomeVFSResult res;
+
+	/* We do support ssh:/// as ssh://localhost/ */
+	hostname = gnome_vfs_uri_get_host_name (uri);
+	if (hostname == NULL) {
+		hostname = "localhost";
+	}
 
 	username = gnome_vfs_uri_get_user_name(uri);
 	if (username == NULL) {
 		username = g_get_user_name();
 	}
-	
+
+	host_port = g_strdup_printf("%d", gnome_vfs_uri_get_host_port(uri) ?
+			gnome_vfs_uri_get_host_port(uri) : 22);
+
 	command_line  = g_strconcat ("ssh -oBatchMode=yes -x -l ", 
 				     username,
-				     " ", gnome_vfs_uri_get_host_name (uri),
-				     " ", "\"LC_ALL=C;", command,"\"",
+				     " -p ", host_port,
+				     " ", hostname,
+				     " ",
+				     "\"LC_ALL=C; echo READY > /dev/stderr;",
+				     command, "; echo DONE > /dev/stderr\"",
 				     NULL);
+	g_free (host_port);
+
+	D(("ssh_connect () command: %s\n", command_line));
 
 	g_shell_parse_argv (command_line, &argc, &argv, &gerror);
+
 	g_free (command_line);
 	if (gerror) {
 		g_warning (gerror->message);
 		return GNOME_VFS_ERROR_BAD_PARAMETERS;
 	}
 
-
-	/* fixme: handle other ports */
 	handle = g_new0 (SshHandle, 1);
 	handle->uri = uri;
 
 	g_spawn_async_with_pipes (NULL, argv, NULL, 
-				  G_SPAWN_SEARCH_PATH | G_SPAWN_STDERR_TO_DEV_NULL,
+				  G_SPAWN_SEARCH_PATH,
 				  NULL, NULL,
 				  &handle->pid, &handle->write_fd, &handle->read_fd,
-				  NULL, &gerror);
+				  &handle->error_fd, &gerror);
 	g_strfreev (argv);
 
 	if (gerror) {
 		g_warning (gerror->message);
 		g_free (handle);
+		return GNOME_VFS_ERROR_GENERIC;
 	}
 
 	gnome_vfs_uri_ref (handle->uri);
 
 	*handle_return = handle;
 
-	return GNOME_VFS_OK;
+	/* You can add more error checking here */
+	memset (buffer, '\0', LINE_LENGTH);
+	res = ssh_read_error (handle->error_fd, &buffer,
+			LINE_LENGTH, &bytes_read);
+
+	if (bytes_read != 0 && res == GNOME_VFS_OK) {
+		if (strncmp ("READY", buffer, strlen ("READY")) == 0) {
+			res = GNOME_VFS_OK;
+		} else if (strncmp ("Permission denied", buffer,
+					strlen("Permission denied")) == 0) {
+			res = GNOME_VFS_ERROR_LOGIN_FAILED;
+		} else if (strncmp ("Host key verification failed", buffer,
+					strlen("Host key verification failed"))
+				== 0) {
+			res = GNOME_VFS_ERROR_SERVICE_NOT_AVAILABLE;
+		} else if (strstr (buffer, "Connection refused") != NULL) {
+			res = GNOME_VFS_ERROR_ACCESS_DENIED;
+		}
+	}
+
+	return res;
 }
 
 static GnomeVFSResult
@@ -207,6 +266,7 @@ ssh_destroy (SshHandle *handle)
 {
 	close (handle->read_fd);
 	close (handle->write_fd);
+	close (handle->error_fd);
 	gnome_vfs_uri_unref (handle->uri);
 	kill (handle->pid, SIGINT);
 	g_free (handle);
@@ -234,6 +294,46 @@ ssh_read (SshHandle *handle,
 	return GNOME_VFS_OK;
 }
 
+static GnomeVFSResult
+ssh_read_error (int error_fd,
+		gpointer buffer,
+		GnomeVFSFileSize num_bytes,
+		GnomeVFSFileSize *bytes_read)
+{
+	GnomeVFSFileSize my_read;
+
+	my_read = (GnomeVFSFileSize) read (error_fd, buffer,
+			(size_t) num_bytes);
+
+	if (my_read == -1) {
+		return gnome_vfs_result_from_errno ();
+	}
+
+	*bytes_read = my_read;
+
+	return GNOME_VFS_OK;
+}
+
+static GnomeVFSResult
+ssh_check_for_done (SshHandle *handle)
+{
+	char buffer[LINE_LENGTH];
+	GnomeVFSResult res;
+	GnomeVFSFileSize bytes_read;
+
+	memset (buffer, '\0', LINE_LENGTH);
+	res = ssh_read_error (handle->error_fd, &buffer,
+			LINE_LENGTH, &bytes_read);
+
+	if (bytes_read != 0 && res == GNOME_VFS_OK) {
+		if (strncmp ("DONE", buffer, strlen ("DONE")) == 0) {
+			return res;
+		}
+	}
+
+	return GNOME_VFS_ERROR_GENERIC;
+}
+
 static GnomeVFSResult
 ssh_write (SshHandle *handle,
 	   gconstpointer buffer,
@@ -262,6 +362,23 @@ ssh_write (SshHandle *handle,
 	return GNOME_VFS_OK;
 }
 
+static GnomeVFSResult
+ssh_wait_and_destroy (SshHandle *handle, GnomeVFSContext *context)
+{
+	int i;
+
+	/* That's 30 seconds */
+	for (i = 0; i < 100 && kill (handle->pid, 0) != -1; i++) {
+		if (gnome_vfs_context_check_cancellation (context) == TRUE) {
+			break;
+		}
+		usleep (SLEEP_TIME);
+	}
+
+	return ssh_destroy (handle);
+}
+
+
 #if 0
 static GnomeVFSResult
 ssh_send (SshHandle *handle,
@@ -300,7 +417,7 @@ do_open (GnomeVFSMethod *method,
 						  G_DIR_SEPARATOR_S);
 		quoted_name = g_shell_quote (name);
 		g_free (name);
-		
+
 		cmd = g_strdup_printf ("cat %s", quoted_name);
 		result = ssh_connect (&handle, uri, cmd);
 		g_free (cmd);
@@ -311,7 +428,7 @@ do_open (GnomeVFSMethod *method,
 	} else {
 		return GNOME_VFS_ERROR_INVALID_OPEN_MODE;
 	}
-	
+
 	handle->open_mode = mode;
 	handle->type = SSH_FILE;
 	*method_handle = (GnomeVFSMethodHandle *)handle;
@@ -319,7 +436,7 @@ do_open (GnomeVFSMethod *method,
 	return GNOME_VFS_OK;
 }
 
-static GnomeVFSResult   
+static GnomeVFSResult
 do_create (GnomeVFSMethod *method,
 	   GnomeVFSMethodHandle **method_handle,
 	   GnomeVFSURI *uri,
@@ -365,7 +482,7 @@ do_close (GnomeVFSMethod *method,
 	  GnomeVFSMethodHandle *method_handle,
 	  GnomeVFSContext *context)
 {
-	return ssh_destroy ((SshHandle *)method_handle);
+	return ssh_wait_and_destroy ((SshHandle *)method_handle, context);
 }
 
 static GnomeVFSResult
@@ -376,12 +493,21 @@ do_read (GnomeVFSMethod *method,
 	 GnomeVFSFileSize *bytes_read,
 	 GnomeVFSContext *context)
 {
-	GnomeVFSResult result;
+	GnomeVFSResult result, res_secondary;
 
 	result =  ssh_read ((SshHandle *)method_handle, buffer, num_bytes,
 			bytes_read);
+	if (result != GNOME_VFS_OK) {
+		return result;
+	}
+
 	if (*bytes_read == 0) {	
-		result = GNOME_VFS_ERROR_EOF;
+		res_secondary = ssh_check_for_done ((SshHandle *)method_handle);
+		if (res_secondary != GNOME_VFS_OK) {
+			result = res_secondary;
+		} else {
+			result = GNOME_VFS_ERROR_EOF;
+		}
 	}
 	return result;
 }
@@ -415,6 +541,10 @@ do_open_directory (GnomeVFSMethod *method,
 	char *quoted_name;
 
 	name = gnome_vfs_unescape_string (uri->text, G_DIR_SEPARATOR_S);
+	if (name == NULL) {
+		return GNOME_VFS_ERROR_NOT_FOUND;
+	}
+
 	quoted_name = g_shell_quote (name);
 
 	if (*name != '\0') {
@@ -431,7 +561,7 @@ do_open_directory (GnomeVFSMethod *method,
 	if (result != GNOME_VFS_OK) {
 		return result;
 	}
-
+	handle->info_opts = options;
 	handle->open_mode = GNOME_VFS_OPEN_NONE;
 	handle->type = SSH_LIST;
 	*method_handle = (GnomeVFSMethodHandle *)handle;
@@ -460,7 +590,6 @@ get_access_info (GnomeVFSURI *uri, GnomeVFSFileInfo *file_info)
      struct param params[2] = {{'r', GNOME_VFS_PERM_ACCESS_READABLE},
                                {'w', GNOME_VFS_PERM_ACCESS_WRITABLE}};
 
-     
      name = gnome_vfs_unescape_string (uri->text, G_DIR_SEPARATOR_S);
 
 
@@ -510,13 +639,18 @@ do_read_directory (GnomeVFSMethod *method,
                    GnomeVFSFileInfo *file_info,
 		   GnomeVFSContext *context)
 {
-	GnomeVFSResult result = GNOME_VFS_OK;
+	GnomeVFSResult result = GNOME_VFS_OK, res_secondary;
 	char line[LINE_LENGTH+1];
 	char c;
 	int i=0;
 	GnomeVFSFileSize bytes_read;
 	struct stat st;
 	char *tempfilename, *filename, *linkname;
+	SshHandle *handle = (SshHandle *)method_handle;
+	const char *mime_type;
+	char *target_name, *parent;
+	GnomeVFSFileInfo *target_info;
+	GnomeVFSURI *target_uri;
 
 	for (;;) {
 		tempfilename = NULL;
@@ -526,14 +660,19 @@ do_read_directory (GnomeVFSMethod *method,
 		bytes_read = 0;
 
 		while (i<LINE_LENGTH) {
-			result = ssh_read ((SshHandle *)method_handle, &c,
+			result = ssh_read (handle, &c,
 					   sizeof(char), &bytes_read);
 			if (bytes_read == 0 || c == '\r' || c == '\n') {
 				break;
 			}
 
 			if (result != GNOME_VFS_OK) {
-				return result;
+				res_secondary = ssh_check_for_done (handle);
+				if (res_secondary != GNOME_VFS_OK) {
+					result = res_secondary;
+				} else {
+					return result;
+				}
 			}
 
 			line[i] = c;
@@ -565,22 +704,60 @@ do_read_directory (GnomeVFSMethod *method,
 		gnome_vfs_stat_to_file_info (file_info, &st);
 		file_info->name = filename;
 		if (linkname) {
+			GNOME_VFS_FILE_INFO_SET_SYMLINK (file_info, TRUE);
+			file_info->valid_fields
+				|= GNOME_VFS_FILE_INFO_FIELDS_SYMLINK_NAME;
 			file_info->symlink_name = linkname;
+			D(("symlink: %s\n", file_info->symlink_name));
 		}
 
-		/* FIXME: support symlinks correctly */
+		/* FIXME: support symlinks to directories correctly */
 
-		file_info->mime_type = g_strdup 
-			(gnome_vfs_get_file_mime_type (filename, &st, FALSE));
+		if (file_info->type == GNOME_VFS_FILE_TYPE_SYMBOLIC_LINK) {
+			parent = gnome_vfs_uri_to_string
+				(handle->uri, 0);
+			target_name =
+				gnome_vfs_make_uri_full_from_relative
+				(parent, file_info->symlink_name);
+
+			if ((handle->info_opts
+				& GNOME_VFS_FILE_INFO_FOLLOW_LINKS) == 0) {
+				mime_type = "x-special/symlink";
+			} else {
+				target_uri = gnome_vfs_uri_new (target_name);
+				mime_type = gnome_vfs_get_file_mime_type
+					(target_name, NULL, FALSE);
+				D(("Looking up mime-type for '%s'\n\n",
+							target_name));
+				target_info = gnome_vfs_file_info_new ();
+				do_get_file_info (method, target_uri,
+						target_info, handle->info_opts,
+						context);
+				file_info->type = target_info->type;
+				gnome_vfs_file_info_unref (target_info);
+				gnome_vfs_uri_unref (target_uri);
+			}
+
+			g_free (target_name);
+			g_free (parent);
+		} else {
+			mime_type = (gnome_vfs_get_file_mime_type
+					(filename, &st, FALSE));
+		}
+
+		if (handle->info_opts & GNOME_VFS_FILE_INFO_GET_MIME_TYPE) {
+			file_info->mime_type = g_strdup (mime_type);
+			file_info->valid_fields
+				|= GNOME_VFS_FILE_INFO_FIELDS_MIME_TYPE;
+		}
 
-		file_info->valid_fields |= GNOME_VFS_FILE_INFO_FIELDS_MIME_TYPE;
 		file_info->valid_fields &= 
 			~GNOME_VFS_FILE_INFO_FIELDS_BLOCK_COUNT;
 		file_info->valid_fields &= 
 			~GNOME_VFS_FILE_INFO_FIELDS_IO_BLOCK_SIZE;
-
-		get_access_info (((SshHandle*)method_handle)->uri, 
-				 file_info);
+		if (handle->info_opts & GNOME_VFS_FILE_INFO_GET_ACCESS_RIGHTS) {
+			get_access_info (handle->uri, file_info);
+		}
 
 		/* Break out.
 		   We are in a loop so we get the first 'ls' line;
@@ -622,7 +799,7 @@ do_get_file_info (GnomeVFSMethod *method,
 	if (result != GNOME_VFS_OK) {
 		return result;
 	}
-
+	handle->info_opts = options;
 	handle->open_mode = GNOME_VFS_OPEN_NONE;
 	handle->type = SSH_LIST;
 
@@ -659,7 +836,7 @@ do_make_directory (GnomeVFSMethod *method,
 		return result;
 	}
 
-	ssh_destroy (handle);
+	ssh_wait_and_destroy (handle, context);
 
 	return result;
 }
@@ -690,18 +867,60 @@ do_remove_directory (GnomeVFSMethod *method,
 		return result;
 	}
 
-	ssh_destroy (handle);
+	ssh_wait_and_destroy (handle, context);
 
 	return result;
 }
 
+static GnomeVFSResult
+do_move (GnomeVFSMethod *method,
+	 GnomeVFSURI *old_uri,
+	 GnomeVFSURI *new_uri,
+	 gboolean force_replace,
+	 GnomeVFSContext *context)
+{
+	SshHandle *handle = NULL;
+	GnomeVFSResult res;
+	gboolean same_fs;
+	char *old_name, *new_name;
+	char *quoted_old_name, *quoted_new_name;
+	char *cmd;
+
+	res = do_check_same_fs (method, old_uri, new_uri, &same_fs, context);
+	if (res != GNOME_VFS_OK) {
+		return res;
+	}
+
+	old_name = gnome_vfs_unescape_string (old_uri->text, G_DIR_SEPARATOR_S);
+	new_name = gnome_vfs_unescape_string (new_uri->text, G_DIR_SEPARATOR_S);
+
+	quoted_old_name = g_shell_quote (old_name);
+	quoted_new_name = g_shell_quote (new_name);
+
+	cmd = g_strdup_printf ("mv %s %s", quoted_old_name, quoted_new_name);
+	res = ssh_connect (&handle, old_uri, cmd);
+	g_free (cmd);
+	g_free (old_name);
+	g_free (new_name);
+	g_free (quoted_old_name);
+	g_free (quoted_new_name);
+
+	if (res != GNOME_VFS_OK) {
+		return res;
+	}
+
+	ssh_wait_and_destroy (handle, context);
+
+	return res;
+}
+
 GnomeVFSResult
 do_unlink (GnomeVFSMethod *method,
 	   GnomeVFSURI *uri,
-	   GnomeVFSContext *contet)
+	   GnomeVFSContext *context)
 {
 	SshHandle *handle = NULL;
-	char *cmd = NULL;
+	char *cmd;
 	GnomeVFSResult result;
 	gchar *name;
 	gchar *quoted_name;
@@ -721,11 +940,46 @@ do_unlink (GnomeVFSMethod *method,
 		return result;
 	}
 
-	ssh_destroy (handle);
+	ssh_wait_and_destroy (handle, context);
 
 	return result;
 }
 
+static GnomeVFSResult
+do_check_same_fs (GnomeVFSMethod  *method,
+		GnomeVFSURI     *a,
+		GnomeVFSURI     *b,
+		gboolean        *same_fs_return,
+		GnomeVFSContext *context)
+{
+	const gchar *a_host_name, *b_host_name;
+	const gchar *a_user_name, *b_user_name;
+
+	g_return_val_if_fail (a != NULL, GNOME_VFS_ERROR_INTERNAL);
+	g_return_val_if_fail (b != NULL, GNOME_VFS_ERROR_INTERNAL);
+
+	a_host_name = gnome_vfs_uri_get_host_name (a);
+	b_host_name = gnome_vfs_uri_get_host_name (b);
+	a_user_name = gnome_vfs_uri_get_user_name (a);
+	b_user_name = gnome_vfs_uri_get_user_name (b);
+
+	g_return_val_if_fail (a_host_name != NULL, GNOME_VFS_ERROR_INVALID_URI);        g_return_val_if_fail (b_host_name != NULL, GNOME_VFS_ERROR_INVALID_URI);                                                                                
+	if (a_user_name == NULL) {
+		a_user_name = g_get_user_name ();
+	}
+	if (b_user_name == NULL) {
+		b_user_name = g_get_user_name ();
+	}
+
+	if (same_fs_return != NULL) {
+		*same_fs_return =
+			((!strcmp (a_host_name, b_host_name))
+			 && (!strcmp (a_user_name, b_user_name)));
+	}
+
+	return GNOME_VFS_OK;
+}
+
 static GnomeVFSResult
 do_set_file_info (GnomeVFSMethod *method,
 		  GnomeVFSURI *uri,
@@ -792,7 +1046,7 @@ do_set_file_info (GnomeVFSMethod *method,
 				break;
 		}
 
-		ssh_destroy (handle);
+		ssh_wait_and_destroy (handle, context);
 	}
 
 	return result;
@@ -808,7 +1062,7 @@ do_get_file_info_from_handle (GnomeVFSMethodHandle *method_handle,
 }
 #endif
 
-gboolean 
+gboolean
 do_is_local (GnomeVFSMethod *method, const GnomeVFSURI *uri)
 {
         return FALSE;
